#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from scipy.stats import kurtosis, skew

def nanskew(a, axis=0, bias=True):
	''' '''
	return skew(a, axis=axis, bias=bias, nan_policy='omit')

def nankurtosis(a, axis=0, fisher=True, bias=True):
	''' '''
	return kurtosis(a, axis=axis, bias=bias, nan_policy='omit')

def nanmin(a, axis=None, out=None):
	''' '''
	try:
		return np.nanmin(a, axis=axis, out=out)
	except:
		return np.nan

def nanmax(a, axis=None, out=None):
	''' '''
	try:
		return np.nanmax(a, axis=axis, out=out)
	except:
		return np.nan

def nanmaxmmin(a, axis=None, out=None):
	''' '''
	return nanmax(a, axis=axis, out=out)-nanmin(a, axis=axis, out=out)

def nanmean(a, axis=None, out=None):
	''' '''
	try:
		return np.nanmean(a, axis=axis, out=out)
	except:
		return np.nan

def nanstd(a, axis=None, dtype=None, out=None):
	''' '''
	try:
		return np.nanstd(a, axis=axis, out=out)
	except:
		return np.nan

def nanmedian(a, axis=None, out=None, overwrite_input=False):
	''' '''
	try:
		return np.nanmedian(a, axis=axis, out=out, overwrite_input=overwrite_input)
	except:
		return np.nan

def q25(x):
	''' '''
	return x.quantile(0.25)

def q75(x):
	''' '''
	return x.quantile(0.75)

def mode(x):
	''' '''
	vc = x.value_counts(normalize=True, sort=True, ascending=False)
	return vc.index[0] if len(vc.index) > 0 else np.nan

def modefreq(x):
	''' '''
	vc = x.value_counts(normalize=True, sort=True, ascending=False)
	return vc.values[0] if len(vc.index) > 0 else np.nan

def lastmode(x):
	''' '''
	vc = x.value_counts(normalize=True, sort=True, ascending=False)
	return vc.index[-1] if len(vc.index) > 0 else np.nan

def lastmodefreq(x):
	''' '''
	vc = x.value_counts(normalize=True, sort=True, ascending=False)
	return vc.values[-1] if len(vc.index) > 0 else np.nan

def nextmode(x):
	''' '''
	vc = x.value_counts(normalize=True, sort=True, ascending=False)
	return vc.index[1] if len(vc.index) > 1 else vc.index[0] if len(vc.index) > 0 else np.nan

def nextmodefreq(x):
	''' '''
	vc = x.value_counts(normalize=True, sort=True, ascending=False)
	return vc.values[1] if len(vc.values) > 1 else vc.values[0] if len(vc.index) > 0 else np.nan



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

from kxy.finance import information_adjusted_correlation as ia_corr

from .base_accessor import BaseAccessor

@pd.api.extensions.register_dataframe_accessor("kxy_finance")
class FinanceAccessor(BaseAccessor):
	"""
	Extension of the pandas.DataFrame class with various finance-specific analytics.

	This class defines the :code:`kxy_finance` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_.

	All its methods defined are accessible from any DataFrame instance as :code:`df.kxy_finance.<method_name>`, so long as the :code:`kxy` python package is imported alongside :code:`pandas`. 
	"""
	def information_adjusted_beta(self, market_column, asset_column, anonymize=False):
		"""
		Estimate the information-adjusted beta of an asset return :math:`r` relative to the market return :math:`r_m`: :math:`\\text{IA-}\\beta := \\text{IA-Corr}\\left(r, r_m \\right) \\sqrt{\\frac{\\text{Var}(r)}{\\text{Var}(r_m)}}`,
		where :math:`\\text{IA-Corr}\\left(r, r_m \\right) := \\text{sgn}\\left(\\text{Corr}\\left(r, r_m \\right) \\right) \\left[1 - e^{-2I(r, r_m)} \\right]` denotes the information-adjusted correlation coefficient, with :math:`\\text{sgn}\\left(\\text{Corr}\\left(r, r_m \\right) \\right)` the sign of the Pearson correlation coefficient.

		Unlike the traditional beta coefficient, namely :math:`\\beta := \\text{Corr}\\left(r, r_m \\right) \\sqrt{\\frac{\\text{Var}(r)}{\\text{Var}(r_m)}}`, that only captures linear relations between market and asset returns, and that is 0 if and only if the two are **decorrelated**, :math:`\\text{IA-}\\beta` captures any relationship between asset return and market return, linear or nonlinear, and is 0 if and only if the two variables are **statistically independent**. 

		Parameters
		----------
		market_column : str
			The name of the column containing market returns.
		asset_column : str
			The name of the column containing asset returns.
		anonymize : bool
			When set to true, your explanatory variables will never be shared with KXY (at no performance cost).


		Returns
		-------
		result : float
			The information-adjusted beta coefficient.

		"""
		assert market_column in self._obj.columns, 'The market column should be a column'
		assert asset_column in self._obj.columns, 'The asset column should be a column'

		m_std = np.nanstd(self._obj[market_column].values)
		a_std = np.nanstd(self._obj[asset_column].values)

		return self.information_adjusted_correlation(market_column, asset_column, anonymize=anonymize)*a_std/m_std



	def information_adjusted_correlation(self, market_column, asset_column, anonymize=False):
		"""
		Estimate the information-adjusted correlation between an asset return :math:`r` and the market return :math:`r_m`: :math:`\\text{IA-Corr}\\left(r, r_m \\right) := \\text{sgn}\\left(\\text{Corr}\\left(r, r_m \\right) \\right) \\left[1 - e^{-2I(r, r_m)} \\right]`, where :math:`\\text{sgn}\\left(\\text{Corr}\\left(r, r_m \\right) \\right)` is the sign of the Pearson correlation coefficient.

		Unlike Pearson's correlation coefficient, which is 0 if and only if asset return and market return are **decorrelated** (i.e. they exhibit no linear relation), information-adjusted correlation is 0 if and only if market and asset returns are **statistically independent** (i.e. the exhibit no relation, linear or nonlinear).


		Parameters
		----------
		market_column : str
			The name of the column containing market returns.
		asset_column : str
			The name of the column containing asset returns.
		anonymize : bool
			When set to true, your explanatory variables will never be shared with KXY (at no performance cost).


		Returns
		-------
		result : float
			The information-adjusted correlation.

		"""
		assert market_column in self._obj.columns, 'The market column should be a column'
		assert asset_column in self._obj.columns, 'The asset column should be a column'

		_obj = self.anonymize(columns_to_exclude=[]) if anonymize else self._obj

		return ia_corr(_obj, market_column, asset_column)










#!/usr/bin/env python
# -*- coding: utf-8 -*-

MODELS_TAKING_FLAT_OUTPUTS = [\
	'sklearn.ensemble.RandomForestRegressor', \
	'sklearn.neural_network.MLPRegressor', 
]

def get_sklearn_learner(class_name, *args, **kwargs):
	'''
	Generate base learner class as a subclass of an sklearn learner class, but one whose hyper-parameters are frozen to user-specified values.

	The class name should be full (e.g. sklearn.ensemble.RandomForestRegressor).
	'''
	import sklearn.ensemble
	import sklearn.gaussian_process
	import sklearn.linear_model
	import sklearn.kernel_ridge
	import sklearn.neural_network
	import sklearn.svm
	import sklearn.tree

	BaseLearner = eval(class_name)
	class Learner(BaseLearner):
		def __init__(self,):
			super(Learner, self).__init__(*args, **kwargs)

		def fit(self, *args, **kwargs):
			args_ = list(args)
			args_[1] = args_[1].flatten() if class_name in MODELS_TAKING_FLAT_OUTPUTS else args_[1]
			return super(Learner, self).fit(*args_, **kwargs)


	return Learner

def get_xgboost_learner(class_name, *args, **kwargs):
	'''
	Generate base learner class as a subclass of an xgboost learner class, but one whose hyper-parameters are frozen to user-specified values.

	The class name should be full (e.g. xgboost.XGBRegressor).
	'''
	import xgboost
	BaseLearner = eval(class_name)
	class Learner(BaseLearner):
		def __init__(self,):
			super(Learner, self).__init__(*args, **kwargs)

	return Learner

# TODO: Add support for tensorflow learners and learners in other frameworks.